package lucastest;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Random;

public class MCMC {


    private final int nSamples;
    private final int sampleEvery;
    private final int seed;
    private final SpeciesSet speciesSet;
    MCMC(Config config, SpeciesSet speciesSet, int nSamples, int sampleEvery) {
        this.nSamples = nSamples;
        this.sampleEvery = sampleEvery;
        this.seed = config.mcmcSeed;
        this.speciesSet = speciesSet;
    }


    void run(GTree gTree, String fName) {
        Random rand = new Random();
        rand.setSeed(seed);
        CalculatorForPosterior posterior = new CalculatorForPosterior();
        GTree oldGTree = gTree.getCopy();
        double oldLogP = posterior.logPosterior(speciesSet, oldGTree);


        try {
            FileWriter fw = new FileWriter(fName);
            fw.write("sample  logP  ");
            fw.write(gTree.logColumnNames());
            fw.write("\n");
            for (int n = 0; n < nSamples; n++) {
                oldGTree = gTree.getCopy();
                doOperator(gTree, rand);
                posterior = new CalculatorForPosterior();
                double newLogP = posterior.logPosterior(speciesSet, gTree);
                double accept = 1.0;
                if (newLogP < oldLogP) {
                    accept = (Math.exp(newLogP - oldLogP));
                }
                if (accept < rand.nextDouble()) {
                    // reject, restore old gtree
                    gTree = oldGTree.getCopy();
                } else {
                    // accept
                    oldLogP = newLogP;
                }
                if ((n % sampleEvery) == 0) {
                    fw.write(String.format("%7d  %8.6f ",n, oldLogP));
                    fw.write(gTree.logColumns());
                    fw.write("\n");
                    int nn = n / sampleEvery;
                    if (nn < 10 || (nn<100 && nn%10==0) || (nn<1000 && nn%100==0) || ( nn%1000==0)) {
                        System.out.println(String.format("%8d logP %8.6f", n, oldLogP));
                    }
                }
            }
            fw.close();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }


    private int randomNonTipNode(GTree newGTree, Random rand) {
        int nNodes = newGTree.getNNodes();
        int nTips = newGTree.getNTips();
        int n = nTips + rand.nextInt(nNodes - nTips);
        return n;
    }


    private void changeNodeHeight(GTree newGTree, Random rand) {
        double y = 0.999*rand.nextDouble();
        double rootRange = 0.002 * (Math.pow(1.0 - y, -2.0) - 1);
        rootRange = Math.min(rootRange, 2.0);
        rootRange = Math.max(rootRange, 0.00001);
        int nNodes = newGTree.getNNodes();
        if (nNodes < 2) {
            return;
        }
        int n = randomNonTipNode(newGTree, rand);
        int anc = newGTree.getNode(n).getAnc();
        int chX = newGTree.getNode(n).getChX();
        int chY = newGTree.getNode(n).getChY();
        double h0 = Math.max(newGTree.getNode(chX).getHeight(), newGTree.getNode(chY).getHeight());
        double newH;
        if  (anc >= 0) {
            double h1 = newGTree.getNode(anc).getHeight();
            newH = h0 + (h1-h0)*rand.nextDouble();
            newGTree.getNode(n).setHeight(newH);
        } else {
            double h = newGTree.getNode(n).getHeight();
            newH = h + rootRange * (rand.nextDouble() - 0.5);
            if (newH < h0) {
                newH = h;
            }
        }
        newGTree.getNode(n).setHeight(newH);

    }

    private void changeSTip(GTree newGTree, Random rand) {
        int n = randomNonTipNode(newGTree, rand);
        int newSTip = rand.nextInt(speciesSet.getNTips());
        newGTree.getNode(n).setSTip(newSTip);
    }



    private int uncleNodeForNNI(GTree newGTree, int n) {
        int unc = -1;
        int anc = newGTree.getNode(n).getAnc();
        if (anc >= 0) {
            int ancanc = newGTree.getNode(anc).getAnc();
            if (ancanc >= 0) {
                int sibX = newGTree.getNode(ancanc).getChX();
                int sibY = newGTree.getNode(ancanc).getChY();
                assert sibY == anc  ||  sibX == anc;
                double ancH = newGTree.getNode(anc).getHeight();
                if (sibX == anc && newGTree.getNode(sibY).getHeight() < ancH) {
                    unc = sibY;
                }
                if (sibY == anc && newGTree.getNode(sibX).getHeight() < ancH) {
                    unc = sibX;
                }
            }
        }
        return unc;
    }


    private void makeSiblingOf(GTree newGTree, int x, int newSib) {
        int anc = newGTree.getNode(newSib).getAnc();
        assert anc >= 0;
        if (newGTree.getNode(anc).getChX() == newSib) {
            newGTree.getNode(anc).setChY(x);
        } else {
            newGTree.getNode(anc).setChX(x);
        }
        newGTree.getNode(x).setAnc(anc);
    }

    /*
    n                  n
     \  sib             \  sib
      \ /   unc          \  \  unc
      anc  /              \  anc
        \ /                \  /
       ancanc             ancanc

      n becomes sibling of anc
      unc becomes sibling of sib
     */
    private void doNNI(GTree newGTree, Random rand) {
        int nNodes = newGTree.getNNodes();
        if (nNodes < 5) {
            return;
        }
        int n = rand.nextInt(nNodes);
        int unc = uncleNodeForNNI(newGTree, n);
        while (unc < 0) {
            n =  rand.nextInt(nNodes);
            unc = uncleNodeForNNI(newGTree, n);
        }
        int anc = newGTree.getNode(n).getAnc();
        int sib = (newGTree.getNode(anc).getChX() == n) ? newGTree.getNode(anc).getChY() : newGTree.getNode(anc).getChX();
        makeSiblingOf(newGTree, n, anc);
        makeSiblingOf(newGTree, unc, sib);
    }




    private void doOperator(GTree newGTree, Random rand) {
        int r = rand.nextInt(3);
        switch (r) {
            case 0: doNNI(newGTree, rand); break;
            case 1: changeNodeHeight(newGTree, rand); break;
            case 2:changeSTip(newGTree, rand); break;
            default: assert false;
        }
    }


}
