package lucastest;


/**
 * Probabilities for a persister k.
 * Contains the info needed to find Pr(k does not coalesce with i or j, given i and j ready to coalesce
 */

public class ProbabilitiesForPersister {
    double PrNk0;
    double PrNk1;
    double PrNik2plusNjk2plus;

    double PrTijkGivenAiAjN000;
    double PrTijkGivenAiAjN001;
    double PrTijkGivenAiAjN010;
    double PrTijkGivenAiAjN100;
    double PrTijkGivenAiAjN110;
    double PrTijkAiAjNik2plusNjk2plus;




    ProbabilitiesForPersister(SpeciesSet speciesSet, ProbabilitiesForCoalescers pfc, Lineage iLin, Lineage jLin, Lineage kLin) {
        assert iLin.endB == jLin.endB;
        assert Utils.floatsVClose(iLin.endT, jLin.endT);
        int b = iLin.endB;
        double t = iLin.endT;
        int kSbIsEb = (kLin.startB == b) ? 1 : 0;
        int kSbIsNotEb = 1 - kSbIsEb;
        int kSbIsiSb = (kLin.startB == iLin.startB) ? 1 : 0;
        int kSbIsNotiSb = 1 - kSbIsiSb;
        int kSbIsjSb = (kLin.startB == jLin.startB) ? 1 : 0;
        int kSbIsNotjSb = 1 - kSbIsjSb;
        double ui = Math.max(0, Math.min(jLin.startT, kLin.startT) - iLin.startT);
        double uj = Math.max(0, Math.min(iLin.startT, kLin.startT) - jLin.startT);
        double uk = Math.max(0, Math.min(iLin.startT, jLin.startT) - kLin.startT);
        double vij = Math.max(0, kLin.startT - Math.max(iLin.startT, jLin.startT));
        double vik = Math.max(0, jLin.startT - Math.max(iLin.startT, kLin.startT));
        double vjk = Math.max(0, iLin.startT - Math.max(jLin.startT, kLin.startT));
        double w = t - Math.max(iLin.startT, Math.max(jLin.startT, kLin.startT));
        double ti = t - iLin.startT;
        double tj = t - jLin.startT;
        double tk = t - kLin.startT;
        assert Utils.floatsVClose(ti, ui+vij+vik+w);
        assert Utils.floatsVClose(tj, uj+vij+vjk+w);
        assert Utils.floatsVClose(tk, uk+vik+vjk+w);
        // fix rounding errors which can make probabilities just over 1
        double vikw = Math.min(vik + w, ti);
        double vjkw = Math.min(vjk + w, tj);
        double uivij = Math.min(ui + vij, ti);
        double ujvij = Math.min(uj + vij, tj);

        double mRate = speciesSet.getMRate();
        double s = speciesSet.getNTips();
        double cRate = speciesSet.getCRate(b);
        double kScRate = speciesSet.getCRate(kLin.startB);

        // probabilities of mig counts.
        double kMT = (s - 1) * mRate * tk;
        PrNk0 = Math.exp(-kMT);
        PrNk1 = kMT * Math.exp(-kMT);
        PrNik2plusNjk2plus = Math.max(0.0, 1.0 - pfc.PrNi0*pfc.PrNj0*PrNk0 - pfc.PrNi0*pfc.PrNj0*PrNk1 -
                pfc.PrNi0*pfc.PrNj1*PrNk0 - pfc.PrNi1*pfc.PrNj0*PrNk0 - pfc.PrNi1*pfc.PrNj1*PrNk0);

        // 6 cases for probability of no i-k or j-k coal, given i and j arrive and the three mig counts
        PrTijkGivenAiAjN000 = PrTijkGivenAiAjN001 = PrTijkGivenAiAjN010 = 0.0;
        PrTijkGivenAiAjN100 = PrTijkGivenAiAjN110 = PrTijkAiAjNik2plusNjk2plus = 0.0;
        if (iLin.startB == b && jLin.startB == b) {
            PrTijkGivenAiAjN000 = kSbIsNotEb + kSbIsEb * Math.exp(-(vikw + vjkw) * cRate);
        }
        if (iLin.startB == b && jLin.startB == b) {
            double x001a = uk * Math.exp(-(vikw + vjkw) * cRate);
            double x001b = (vik + vjk) * Utils.uexp((ui + uj) * cRate) * Math.exp(-2 * w * cRate);
            double x001c = w * Utils.uexp(2 * w * cRate);
            PrTijkGivenAiAjN001 = kSbIsEb + kSbIsNotEb * (1 / (s - 1)) * (1 / tk) * (x001a + x001b + x001c);
        }
        if (iLin.startB == b &&  jLin.startB != b) {
            double x010a = kSbIsNotEb * kSbIsNotjSb;
            double x010bi = kSbIsEb * kSbIsNotjSb * Math.exp(-vikw * cRate);
            double x010bj = (1 / tj) * (ujvij * Math.exp(-vjkw * cRate) + vjkw * Utils.uexp(vjkw * cRate));
            double x010b = x010bi * x010bj;
            double x010c = kSbIsNotEb * kSbIsjSb * (1 / tj) * (ujvij + vjkw * Utils.uexp(vjkw * kScRate));
            assert kSbIsEb * kSbIsjSb != 1;
            PrTijkGivenAiAjN010 = x010a + x010b + x010c;
        }
        if (iLin.startB != b &&  jLin.startB == b) {
            double x100a = kSbIsNotEb * kSbIsNotiSb;
            double x100bj = kSbIsEb * kSbIsNotiSb * Math.exp(-vjkw * cRate);
            double x100bi = (1 / ti) * (uivij * Math.exp(-vikw * cRate) + vikw * Utils.uexp(vikw * cRate));
            double x100b = x100bj * x100bi;
            double x100c = kSbIsNotEb * kSbIsiSb * (1 / ti) * (uivij + vikw * Utils.uexp(vikw * kScRate));
            assert kSbIsEb * kSbIsiSb != 1;
            PrTijkGivenAiAjN100 = x100a + x100b + x100c;
        }
        if (iLin.startB != b &&  jLin.startB != b) {
            double x110ai = (1/ti) * (uivij * Math.exp(-vikw * cRate) + vikw * Utils.uexp(vikw * cRate));
            double x110aj = (1/tj) * (ujvij * Math.exp(-vjkw * cRate) + vjkw * Utils.uexp(vjkw * cRate));
            double x110a = kSbIsEb * x110ai * x110aj;
            double x110bi = kSbIsNotiSb + kSbIsiSb * (1/ti) * (uivij + vikw * Utils.uexp(vikw * kScRate));
            double x110bj = kSbIsNotjSb + kSbIsjSb * (1/tj) * (ujvij + vjkw * Utils.uexp(vjkw * kScRate));
            double x110b = kSbIsNotEb * x110bi * x110bj;
            PrTijkGivenAiAjN110 = x110a + x110b;
        }

        {
            double prN2 = pfc.PrAi * pfc.PrAj * PrNik2plusNjk2plus;
            MigrationModel migrationModel = new MigrationModel(s, mRate);
            ThreeStateCTAModel threeStateCTAModel = new ThreeStateCTAModel(s, mRate, pfc.speciesSet.getMeanCRate());
            {
                int iSbIskSb = (iLin.startB == kLin.startB) ? 1 : 0;
                double CTAInitialPr2ik = migrationModel.getPrInBranch(iSbIskSb, ui + vij + uk + vjk);
                double CTAPr2or3ik = 0;
                CTAPr2or3ik += threeStateCTAModel.getCTAPr2Given3(w) * (1 - CTAInitialPr2ik);
                CTAPr2or3ik += threeStateCTAModel.getCTAPr2Given2(w) * CTAInitialPr2ik;
                CTAPr2or3ik += threeStateCTAModel.getCTAPr3Given3(w) * (1 - CTAInitialPr2ik);
                CTAPr2or3ik += threeStateCTAModel.getCTAPr3Given2(w) * CTAInitialPr2ik;
                prN2 *= CTAPr2or3ik;
            }
            {
                int jSbIskSb = (jLin.startB == kLin.startB) ? 1 : 0;
                double CTAInitialPr2jk = migrationModel.getPrInBranch(jSbIskSb, uj + vij + uk + vik);
                double CTAPr2or3jk = 0;
                CTAPr2or3jk += threeStateCTAModel.getCTAPr2Given3(w) * (1 - CTAInitialPr2jk);
                CTAPr2or3jk += threeStateCTAModel.getCTAPr2Given2(w) * CTAInitialPr2jk;
                CTAPr2or3jk += threeStateCTAModel.getCTAPr3Given3(w) * (1 - CTAInitialPr2jk);
                CTAPr2or3jk += threeStateCTAModel.getCTAPr3Given2(w) * CTAInitialPr2jk;
                prN2 *= CTAPr2or3jk;
                PrTijkAiAjNik2plusNjk2plus = prN2;
            }
        }
        checkValidity();
    }


    public String asText() {
        StringBuffer buffer = new StringBuffer();

        buffer.append(String.format("PrNk0 %s PrNk1 %s ", Utils.nonNegIn8Chars(PrNk0), Utils.nonNegIn8Chars(PrNk1)));
        buffer.append(String.format("PrNik2plusNjk2plus %s \n", Utils.nonNegIn8Chars(PrNik2plusNjk2plus)));
        buffer.append(String.format("PrTijkGivenAiAjN000 %s PrTijkGivenAiAjN001 %s ", Utils.nonNegIn8Chars(PrTijkGivenAiAjN000), Utils.nonNegIn8Chars(PrTijkGivenAiAjN001)));
        buffer.append(String.format("PrTijkGivenAiAjN010 %s PrTijkGivenAiAjN100 %s ", Utils.nonNegIn8Chars(PrTijkGivenAiAjN010), Utils.nonNegIn8Chars(PrTijkGivenAiAjN100)));
        buffer.append(String.format("PrTijkGivenAiAjN110 %s PrTijkGivenAiAjNik2plusNjk2plus %s \n", Utils.nonNegIn8Chars(PrTijkGivenAiAjN110), Utils.nonNegIn8Chars(PrTijkAiAjNik2plusNjk2plus)));
        return buffer.toString();
    }



    private void checkValidity() {
        assert Utils.validProbability(PrNk0);
        assert Utils.validProbability(PrNk1);
        assert Utils.validProbability(PrNik2plusNjk2plus);
        assert Utils.validProbability(PrTijkGivenAiAjN000);
        assert Utils.validProbability(PrTijkGivenAiAjN001);
        assert Utils.validProbability(PrTijkGivenAiAjN010);
        assert Utils.validProbability(PrTijkGivenAiAjN100);
        assert Utils.validProbability(PrTijkGivenAiAjN110);
        assert Utils.validProbability(PrTijkAiAjNik2plusNjk2plus);
    }


}
